/*
** c99 compiler for the TI99/4A
**
** (c) 1985, 1987 by Clint Pulley
**
** Edit # 167  : 25-Oct-87  1110
** Edit # ???  : 28-Dec-99  0100  (Kristine Rogers)
**
** Based on Jeff Lomicka's version (82/01/25)
** of the small-c compiler by Ron Cain
**
** Extensions based on the work of Jim Hendrix
*/
/*	c99 Update History
Version  Date    Changes and new features:
  1.3  85/09/09  -Original release version.
  1.31 85/11/24  -Improves code for ++ and -- operators.
  1.32 85/12/01  -Implements \ completely.
                 -Allows quoted filenames in #include.
  1.4  85/12/11  -Improves generated code by eliminating pop from R9 whenever
                  possible (revised CSUP).
                 -Generates DEF MAIN automatically.
                 -Displays errors on screen and reports only the first error
                  a statement.
                 -Has default screen colors of white/blue.
  1.41 85/12/22  -Corrects problems with pointer arithmetic & logical operators.
                 -Changes compiler screen display.
                 -Outputs literals after each function.
                 -Partially implements , operator.
  1.5  86/01/05  -Allows local declarations only at start of a block.
                 -Parses argument declarations of the form (*func)().
                 -Implements do/while.
  1.51 86/01/15  -Converts character type test and string compare functions to
                  assembly code for memory reduction and faster compiles.
                 -Allows _ in names.
  1.6  86/01/19  -Implements switch/case (revised CSUP).
                 -Fully implements , operator.
  1.7  86/01/26  -Implements for.
                 -Removes #outseg.
  1.8  86/02/01  -Implements initializers on global declarations.
  1.9  86/02/01  -Provides inline push code option.
  2.0  86/02/02  -Provides descriptive error messages on screen.
  2.1x 86/08/08  -Implements 2 dimensional arrays.
                 -Implements extern and entry.
  2.1  86/10/14  -Corrects problem with generated labels for initializers when
                  c-text included.
  2.2  87/01/18  -Implements conditional compile directives.
                 -Expands static symbol table to 256 entries.
  2.3  87/03/05  -Implements &&, || with early dropout.
                 -Eliminates use 0f 99/4A Pad addresses in generated code.
                 -Uses hash coding for global symbol table access.
  3.0  87/04/18  -Eliminates generation of ABS instruction before conditionals.
                 -Corrects problem with Pad address elimination.
                 -Treats while(1) as a special case.
                 -Revises indirect call to not use R10.
                 -Changes code for byte (char) load/store using R10 as pointer
                  to ls byte of R8.
                 -Requires revised CSUP.
  3.1  87/10/10  -Implements pointer arrays.
                 -Implements conditional expression operator.
  3.2  87/10/10  -Implements <gasp!> goto.
                 -Reports type of missing token.
  3.3  87/10/13  -Corrects pointer array implementation.
  4.0  87/10/24  -Final 99/4A version.
                 -Reports lines processed, global and local symbol counts.
                 -Implements hex constants.
                 -Issues "call files(3)" during initialization.
                 -Corrects problem with decimal input.
*/

/* >>>>>>> start of c99c1 <<<<<<<< 
**
** (c) 1987 by Clint Pulley
**
** Last Edit 87/03/05  2300
** Edit # ???  : 28-Dec-99  0100  (Kristine Rogers)
*/
/*
** Compiler begins execution here
*/

#include <stdio.h>
#include <ctype.h>

#pragma inline_depth(250)
#pragma inline_recursion(on)

#define	C99_MAIN

#include "c99.h"

#undef C99_MAIN

void main(int argc, char* argv[])
{
	short helpexit = FALSE;

	/*	Clear global symbols	*/
	for (cptr = 0; cptr < GSYMSZ; cptr++)
	{
		g_sym_tab[cptr] = '\0';
	}

	/*	clear macros	*/
	for (cptr = 0; cptr < MACQSIZE; cptr++)
	{
		macro_pool[cptr] = '\0';
	}

	/*	Clear local symbols	*/
	for (cptr = 0; cptr < LSYMSZ; cptr++)
	{
		l_sym_tab[cptr] = '\0';
	}

	locptr = SYMSIZ;	/*	clear local symbols		*/
	test_mode = FALSE;
	macptr = MACQ;		/*	Clear macro pool		*/
	wqindex = 0;		/*	Clear "while" queue		*/
	swnext = swq;		/*	Clear switch table		*/

	/*	...all set to zero or NO ...	*/
	swactive = FALSE;	/*	Not inside "switch"		*/
	litptr = 0;			/*	Clear literal pool		*/
	iflv = 0;			/*	Clear conditional compile	*/
	skiplv = 0;
	pmode = FALSE;		/*	Not in preprocessor		*/
	stkp  = 0;			/*	Stack ptr (relative)	*/
	numlin = 0;			/*	# source lines			*/
	numglbs = 0;		/*	# global symbols used	*/
	numlocs = 0;		/*	# local symbols used	*/
	errcnt = 0;			/*	# errors				*/
	errflg = 0;			/*	Not skipping errors		*/
	eof = FALSE;		/*	Not EOF yet				*/
	input = 0;			/*	No input file			*/
	input2 = 0;			/*	 or include file		*/
	output = 0;			/*	No open units			*/
	ncmp = 0;			/*	No open compound states	*/
	nxtlab = 0;			/*	Set label #				*/

	cmode = TRUE;		/*	Enable preprocessing	*/
	decl = -1;			/*	Inhibit local declarations	*/

	/*
	**	Compiler body
	*/
	base_dir[0] = '\0';

	helpexit = ask_opts(argc, argv);		/*	get user options		*/

	if (!helpexit)
	{
		input = getfn(TRUE, infile, "r");	/*	Get initial input file	*/
		output = getfn(FALSE, outfile, "w");	/*  Get initial output file */

		header();			/*	Intro (startup) code	*/
		preprocess();		/*	Start input				*/
		parse();			/*	Process ALL input		*/
		trailer();			/*	Follow-up (exit) code	*/

		if (ctext)
		{
			sumstats();		/*	Add summary and stats to output file	*/
		}

		FCLOSE(output);	/*	Close output file	*/

		output = 0;

		summary();			/*	Summarize compilation	*/
	}

	/*	Then exit to system	*/
}

/*
** Process all input text
*/
/* At this level, only static declarations,
** defines, includes, and function
** definitions are legal...
*/
void parse(void)
{
	while (!eof)				/*	Do until no more input	*/
	{ 
		if (amatch("extern", 6))
		{						/*	External declaration	*/
			dodeclare(EXTERNAL);
			blanks();			/*	Force EOF if pending	*/
			continue;
		}

		if (dodeclare(STATIC))
		{						/*	Static declaration	*/
			blanks();			/*	Force EOF if pending	*/
			continue;
		}

		if (amatch("entry", 5))
		{
			doentry();
			ns();
			blanks();			/*	Force EOF if pending	*/
			continue;
		}

		if (match("#asm"))
		{						/*	assembly-code, just		*/
			doasm();			/*	insert into code.		*/
			blanks();			/*	Force EOF if pending	*/
			continue;
		}

		if (match("#includ"))
		{						/*	Include-file	*/
			doinclude();
			blanks();			/*	Force EOF if pending	*/
			continue;
		}

		if (match("#define"))
		{
			addmac();
			blanks();			/*	Force EOF if pending	*/
			continue;
		}

		newfunc();
		blanks();				/*	Force EOF if pending	*/
	}
}

/*
** Dump the symbol-table
*/
void dumpglbsyms(void)
{
	if (ctext)
	{
		if (output)
		{
			/*	Output file is open; add header.	*/
			outcom(" ");
			outcom(" Global Symbols");
			outcom(" ");
		}
		else
		{
			/*	Output file is not open; print header.	*/
			ol("\n Global Symbols\n");
		}

		dumpsyms(STARTGLB, ENDGLB);
	}
}


/*
** Dump the local symbol-table
*/
void dumplocsyms(void)
{
	if (ctext)
	{
		if (output)
		{
			outcom(" ");
			outcom(" Local Symbols");
			outcom(" ");
		}
		else
		{
			ol("\n Local Symbols\n");
		}

		dumpsyms(STARTLOC, ENDLOC);
	}
}

/*********
Dump a symbol-table
**********/
void dumpsyms(char *StartAddr, char *EndAddr)
{
	short count = 0;
	short id = 0;
	short type = 0;
	short offset = 0;
	short size = 0;
	short symbol = 0;
	short symcount = 0;
	short pointer = 0;

	/*	Print header for Symbol-table	*/
	while (&StartAddr[symbol] < EndAddr)
	{
		/*	For each valid table-entry...	*/
		if (alnum(StartAddr[symbol]))
		{
			symcount++;
		}

		symbol += SYMSIZ;
	}

	if (!symcount)
	{
		if (output)
		{
			outcom("    None used.");
			outcom(" ");
		}
		else
		{
			ol("    None used.\n");
		}

		return;
	}

	if (output)
	{
		outcom("    Symbol   RefersTo   Type  Offset   Size ");
		outcom(" ");
	}
	else
	{
		ol("   Symbol   RefersTo   Type  Offset   Size \n");
	}

	symbol = 0;

	while (&StartAddr [symbol] < EndAddr)
	{
		/*	For each valid table-entry...	*/
		if (alnum(StartAddr[symbol]))
		{
			count = 0;
			pointer = symbol;

			if (output)
			{
				comment();
			}

			outstr("    ");

			while (StartAddr[pointer] != '\0')
			{
				count++;

				outbyte(StartAddr[pointer]);

				pointer++;
			}

			while (count < 9)
			{
				count++;
				outbyte(SPACE);
			}

			id = (short)StartAddr[symbol + IDENT];
			id &= 0xff;

			outstr(SymTypes[id]);

			if (id != FUNCTION)
			{
				type = (short)StartAddr[symbol + TYPE];
				type &= 0xff;

				outstr("   ");

				outstr(VarTypes[type]);
				offset = (unsigned short)getint(&StartAddr[symbol + OFFSET]);
				offset &= 0xffff;

				outstr("   ");

				count = outdec(offset);

				while (count < 6)
				{
					count++;
					outbyte(SPACE);
				}

				size = getint(&StartAddr[symbol + ROWSIZ]);
				size &= 0xffff;

				outstr("  ");

				outhex(size);
			}

			outbyte(LINEFEED);
		}

		symbol += SYMSIZ;
	}
}

/*
** Dump the macro-table
*/
void dumpmacs()
{
	short count = 0;
	short mac_count = 0;
	short index = 0;

	if (!ctext)
	{
		return;
	}

	if (output)
	{
		outcom(" ");
		outcom(" Macros");
		outcom(" ");
		outcom("    Name     Definition ");
		outcom(" ");
	}
	else
	{
		ol("\n Macros\n");
		ol("   Name     Definition \n");
	}

	mac_count = 0;
	index = 0;

	while (index < MACQSIZE)
	{ 
		count = 0;

		if (output)
		{
			comment();
		}

		outstr ("    ");

		while (macro_pool[index] != '\0')
		{
			count++;

			outbyte(macro_pool[index]);
			index++;
		}

		index++;

		while (count < 9)
		{			
			outbyte(SPACE);
			count++;
		}

		while (macro_pool[index] != '\0')
		{
			outbyte(macro_pool[index]);
			index++;
		}

		index++;

		outbyte(LINEFEED);

		if (macro_pool[index] == '\0')
		{
			break;
		}
	}
}

/*
** Dump the literal pool
*/
void dumplits(short typ)
{
	short j = 0;
	short index = 0;

	while (index < litptr) /* to loop with */
	{ 
		if (typ == CCHAR)
		{ 
			ot("BYTE");
			tab();
			j = 12;
		}
		else
		{ 
			ddata();
			j = 8;
		}

		while (j--)
		{ 
			outdec(getlit(index, typ));

			index = index + typ;

			if (!j || (index >= litptr))
			{ 
				newline();    /* need <cr> */
				break;
			}
			outbyte(',');  /* item separator */
		}
	}
}

/*
** Report at end of file
*/
void sumstats()
{
	/* see if anything left hanging... */
	if (ncmp)
	{
		errrpt("Missing closing brace '}'.");
	}

	if (!output)
	{
		return;
	}

	outcom(" ");
	outstr("*   ");
	outdec(numlin);

	ol("source line(s) processed");

	outcom(" ");
	outstr("*   ");
	outdec(numglbs);

	ol("global symbol(s) used");

	outstr("*   ");
	outdec(numlocs);

	ol("(max) local symbols used");

	outcom(" ");

	outstr("   ");
	outdec(errcnt);
	ol("error(s) found.");
	outcom (" ");

	dumpglbsyms ();
	dumpmacs ();
}

/*
** Report for user
*/
void summary()
{
	/* see if anything left hanging... */
	if (ncmp)
	{
		errrpt("Missing '}'.");
	}

	pl("\n\n***** c99 Compilation Complete *****\n\n");

	outdec(numlin);
	ol("source line(s) processed\n");

	outdec(numglbs);
	ol("global symbols used");

	outdec(numlocs);
	ol("(max) local symbols used\n");

	outdec(errcnt);  /* total # errors */
	ol("error(s) found.\n");
}

BOOL get_opt(char* prompt)
{
	PUTS(prompt);
	tigets(line, LINEMAX);

	return isy();
}

/*
** Get options from user
*/
short ask_opts(int argc, char* argv[])
{
	int index = 0;
	char* param = NULL;
	char buffer[80] = { '\0' };
	char* yes = " [YES]\n";
	char* no = " [NO]\n";

	kill();    /* clear input line */

	pl(logo);  /* print banner */
	newline();

	/*	Set defaults to FALSE	*/
	atext = FALSE;
	rtext = FALSE;
	ctext = FALSE;
	pushin = FALSE;
	test_mode = FALSE;
	
	for (index = 1; index < argc; index++)
	{
		param = argv[index];

		if (*param == '-')
		{
			switch(*(param + 1))
			{
				case 'c':
				case 'C':
					ctext = TRUE;
					break;

				case 'a':
				case 'A':
					atext = TRUE;
					break;

				case 'r':
				case 'R':
					rtext = TRUE;
					break;

				case 'p':
				case 'P':
					pushin = TRUE;
					break;

				case 't':
				case 'T':
					test_mode = TRUE;
					break;

				case 'i':
				case 'I':
					strcpy(infile, param + 2);
					break;

				case 'o':
				case 'O':
					strcpy(outfile, param + 2);
					break;

				case '?':
					PUTS("\n* Usage: smallc990 [options] -i[inputfile] -o[outputfile]\n\n");
					PUTS("\t-c\tInclude c-text with output.\n");
					PUTS("\t-a\tInclude assembly-language comments.\n");
					PUTS("\t-r\tInclude 'R' prefix with register numbers.\n");
					PUTS("\t-p\tUse in-line push code.\n");
					PUTS("\t-t\tRun compiler in test mode.\n");
					PUTS("\t-?\tThis help text.\n\n");
					return TRUE;
					break;

				default:
					break;
			}
		}
	}

	PUTS("\n- Including c-text");
	PUTS(ctext ? yes : no);

	PUTS("- Adding assembly-language comments");
	PUTS(atext ? yes : no);

	PUTS("- Using 'R' to prefix register numbers");
	PUTS(rtext ? yes : no);

	PUTS("- Using in-line push code");
	PUTS(pushin ? yes : no);

	PUTS("- Running compiler in test mode");
	PUTS(test_mode ? yes : no);
	PUTS("\n");

	rstr = rtext ? "R" : "";

	kill();		/* Erase line */

	return FALSE;
}

/*
** Open an include file
*/
void doinclude(void)
{
	char* cp = NULL;

	blanks();  /* skip over to name */

	cp = &line[lptr];

	if (*cp == '"')
	{ 
		++lptr;

		while (*++cp != '"')
		{
			;
		}

		*cp = '\0';
	}

	input2 = FOPEN(line + lptr, "r");

	if (!input2)
	{
		errrpt("Invalid filename.");
	}

	kill();    /* clear rest of line */
	/* so next read will come from new file (if open */
}

/* test for global declarations
*/
short dodeclare(short class)
{
	if (amatch("char", 4))
	{ 
		declglb(CCHAR, class);
		ns();

		return TRUE;
	}
	else if (amatch("int", 3) || (class == EXTERNAL))
	{ 
		declglb(CINT, class);
		ns();

		return TRUE;
	}

	return FALSE;
}

/*
** Declare a static variable
**   (i.e. define for use)
*/
/* makes an entry in the symbol table so subsequent
** references can call symbol by name */
void declglb(short type, short class)   /* typ is CCHAR or CINT */
{
	short k = 0;
	short j = 0;
	short r = 0;

	do
	{ 
		r = 0;

		if (endst())
		{
			break;  /* do line */
		}

		if (match ("*") || match("(*"))  /* pointer ? */
		{ 
			j = POINTER; 		/*	yes	*/
			k = 0; 
		}
		else
		{ 
			j = VARIABLE; 		/*	no	*/
			k = 1; 
		}

		if (!symname(ssname))	/* name ok? */
		{
			illname();		/* no... */
		}

		if (findglb(ssname)) /* already there? */
		{
			multidef();
		}

		match(")");

		if (match("()"))
		{
			j = FUNCTION;
			k = FUNCTION;
		}
		else
		{
			if (match ("["))		/* array? */
			{ 
				k = needsub();
				r = cdim2();

				k = k * r;		/* !0 = array */

				if (j == POINTER)
				{
					j = PTRAR1;
				}
				else
				{
					j = ARRAY;
				}
			}
		}

		if (class == EXTERNAL)
		{ 
			ot("REF");
			tab();
			otlbll(ssname);
		}
		else if (j != FUNCTION)
		{ 
			j = initials(type, j, k);

			if (type == CCHAR)
			{
				ol("EVEN");
			}
		}

		addglb(ssname, j, type, k, r); /* add symbol */

	}  while (match (","));
}

/*
** Declare local variables
** (i.e. define for use)   */
/*
** adds symbol table entry with appropriate
** stack offset to find it again */
void declloc(short type)		/* typ is CCHAR or CINT */

{
	short k = 0;
	short j = 0;
	short t = 0;
	short r = 0;

	if (swactive || noloc || (decl < 0))
	{
		oocerr();	/* illegal in switch or with goto */
	}

	do
	{ 
		r = 0;

		if (endst())
		{
			return;
		}

		k = 2;

		j = match("*") ? POINTER : VARIABLE;

		if (!symname(ssname))
		{
			illname();
		}

		if (findloc(ssname))
		{
			multidef();
		}

		if (match("["))
		{
			t = needsub();

			if (t)
			{
				/*	Dimension was specified; use it.	*/
				j = ((j == POINTER) ? PTRAR1 : ARRAY);

				r = cdim2();	/*	check for second dimension.	*/
				t = t * r;

				if ((type == CINT) || (j == PTRAR1))
				{
					k = t + t;
				}
				else
				{
					k = t + (short)(t & 1);
				}
			}
			else
			{
				/*	Dimension was NOT specified.	*/
				j = ((j == POINTER) ? j = PTRAR2 : POINTER);
			}
		}

		decl = decl + k;
		addloc(ssname, j, type, stkp - decl, r);

	}  while (match (","));
}

/*
** Get required array size
*/
/* invoked when declared variable is followed by "["
** this routine makes subscript the absolute
** size of the array.
*/
short needsub(void)
{
	short nsval = 0;	

	if (match("]"))
	{
		return 0;  /* NULL size */
	}

	if (!number(&nsval))  /* go after a number */
	{ 
		ntcerr();		/*	it isn't		*/
		nsval = 1;		/*	so force one	*/
	}

	if (nsval < 0)
	{ 
		errrpt("Illegal subscript value.");
		nsval = (-nsval);
	}

	needbrack("]");		/* force single dimension */

	return nsval;		/* and return size */
}


/*	Generate DEF directives for entry points	*/
void doentry(void)
{
	do
	{ 
		if (endst())
		{
			return;
		}

		if (!symname(ssname))
		{
			illname ();
		}

		ot("DEF");
		tab();
		otlbll(ssname);

	} while (match (","));
}


/*
** Check for second dimension
*/
/* return subscript value if found
** return 1 if not found or error
*/
short cdim2(void)
{
	short r = 0;

	if (match ("["))
	{
		r = needsub();

		if (r)
		{
			return r;
		}
		else
		{
			errrpt("Array dimension required.");
		}
	}

	return 1;
}


/*
** Begin a function
*/
/* Called from "parse" this routine tries to make a function
** out of what follows.
*/
void newfunc(void)
{
	short my_ptr = 0;

	lastst = 0;
	litptr = 0;
	nogo = 0;
	noloc = 0;
	numltf = 0;

	/*	Ignore return-type; assume it ALWAYS short	*/
	if (match("void") || match("int") || match("char"))
	{
		match("*");
	}

	/*	Get next slot in label table.	*/
	litlab = getlabel();

	if (!symname(ssname))
	{
		illname();
		kill();		/* invalidate line */

		return;
	}

	my_ptr = findglb(ssname);

	if (my_ptr)		/*	already in symbol table?	*/
	{
		if (g_sym_tab[my_ptr + IDENT] != FUNCTION)
		{
			multidef();
		}
		else
		{
			if (g_sym_tab[my_ptr + OFFSET] == FUNCTION)
			{
				multidef ();
			}
			else
			{
				g_sym_tab[my_ptr + OFFSET] = FUNCTION;
			}
		}
	}
	else
	{
		addglb(ssname, FUNCTION, CINT, FUNCTION, 1);
	}

	/*	we had better see open paren for args...	*/
	if (!match("("))
	{
		errrpt("Missing open parenthesis '('.");
	}

	if (astreq(ssname, "main", 4))
	{
		ol("DEF MAIN");
	}

	otlbll(ssname); 
	pl(ssname); /* print function name */

	/* ** Clear local stack pointer */
	locptr = SYMSIZ;
	argstk = 0;    /* init arg count */

	while (!match(")"))  /* then count args */
	{
		if (match("void"))
		{
			match ("*");

			if (!alnum(cur_ch()))
			{
				continue;
			}
		}

		if (symname(ssname))
		{ 
			if (findloc(ssname))
			{
				multidef();
			}
			else
			{  
				addloc(ssname, LABEL, LABEL, argstk, 1);
				argstk = (short)(argstk + 2);
			}
		}
		else
		{
			illname();
			junk();
		}

		blanks();

		/* if not closing paren, should be comma */
		if (!streq(")"))
		{
			if (!match(","))
			{
				ndcerr();
			}
		}

		if (endst())
		{
			break;
		}
	}

	stkp = 0;      /* preset stack ptr */
	argtop = argstk;

	while (argstk)
	{
		if (amatch("char", 4))
		{
			getarg(CCHAR);
			ns();
		}
		else if (amatch("int", 3))
		{
			getarg(CINT);
			ns();
		}
		else
		{
			errrpt("Missing arguments."); 
			break; 
		}
	}

	statement();

	if ((lastst != STRETURN) && (lastst != STGOTO))
	{
		ret();
	}

	if (litptr)
	{ 
		printlabel(litlab);
		dumplits(CCHAR);
		ol("EVEN");
	}

	if (numltf > numlocs)
	{
		numlocs = numltf;
	}
}

/*
** Declare argument types
*/
/* called from "newfunc" this routine adds an entry in the
** local symbol table for each named argument
** Function completely rewritten  rev. 1P */
void getarg(short t)    /* t = CCHAR or CINT */
{
	short argptr = 0;
	BOOL legalname = FALSE;
	short j = 0;
	short r = 0;
	short data = 0;

	for ( ; ; )
	{ 
		if (!argstk)
		{
			return;  /* no more args */
		}

		if (match ("*") || match ("(*"))
		{
			j = POINTER;
		}
		else
		{
			j = VARIABLE;
		}

		legalname = symname(ssname);

		if (!legalname)
		{
			illname();
		}

		match (")");	/*	parse (*fcn) ()	*/
		match ("()");

		if (match ("["))		/* pointer? */			
		{
			/*	it is a pointer, so skip all between "[]"	*/
			while (inbyte () != ']')
			{
				if (endst())
				{
					break;
				}
			}

			if (j == POINTER)
			{
				j = PTRAR2;
			}
			else
			{
				j = POINTER;
			}

			r = cdim2();
		}

		if (j == VARIABLE)
		{
			t = CINT;
		}

		if (legalname)
		{
			argptr = findloc(ssname);

			if (argptr)
			{
				/* Add in details of the type and address of */
				/* the name  */
				l_sym_tab[argptr + IDENT] = (char)j;
				l_sym_tab[argptr + TYPE] = (char)t;

				data = (short)getint(&l_sym_tab[argptr + OFFSET]);
				data = argtop - data;

				putint(data, &l_sym_tab[argptr + OFFSET]);
/* 				putint(argtop - getint(argptr + OFFSET), argptr + OFFSET); */
				putint(r, &l_sym_tab[argptr + ROWSIZ]);
			}
			else
			{
				errrpt("Not a valid argument.");
			}
		}

		argstk = (short)(argstk - 2);  /* cnt down */

		if (endst())
		{
			return;
		}

		if (!match (","))
		{
			ndcerr();
		}
	}
}


/*
** Statement parser
*/
/* called whenever syntax requires
** a statement.
** this routine performs that statement
** and returns a number telling which one */
short statement(void)
{
	if (!cur_ch() && eof)
	{
		return 0;
	}

	if (amatch("char", 4))
	{ 
		declloc(CCHAR);
		ns();

		return lastst;
	}

	if (amatch("int", 3))
	{ 
		declloc(CINT);
		ns();

		return lastst;
	}

	if (decl >= 0)
	{ 
		stkp = modstk(stkp - decl);

		if (ncmp > 1)
		{
			nogo = decl; /* disable goto in inner blocks */
		}

		decl = -1;
	}

	if (match("{"))		/*	delete this (})	*/
	{
		compound();
		return lastst;
	}

	if (amatch("if", 2))
	{ 
		doif();
		lastst = STIF;
		return lastst;
	}

	if (amatch("while", 5))
	{ 
		dowhile();
		lastst = STWHILE;
		return lastst;
	}

	if (amatch("do", 2))
	{ 
		dodo();
		lastst = STDO;
		return lastst;
	}

	if (amatch("for", 3))
	{ 
		dofor(); 
		lastst = STFOR;
		return lastst;
	}

	if (amatch("switch", 6))
	{
		doswitch();
		lastst = STSWITCH;
		return lastst;
	}

	if (amatch("case", 4))
	{ 
		docase(); 
		lastst = STCASE;
		return lastst;
	}

	if (amatch("defaul", 6))
	{
		dodefault(); 
		lastst = STDEF;
		return lastst;
	}

	if (amatch("goto", 4))
	{ 
		dogoto(); 
		lastst = STGOTO;
		return lastst;
	}

	if (dolabel())
	{
		lastst = STLABEL;
		return lastst;
	}

	if (amatch("return", 6))
	{ 
		doreturn();
		ns();
		lastst = STRETURN;
		return lastst;
	}
 
	if (amatch("break", 5))
	{ 
		dobreak();
		ns();
		lastst = STBREAK;
		return lastst;
	}

	if (amatch("contin", 6))
	{ 
		docont();
		ns();
		lastst = STCONT;
		return lastst;
	}

	if (match(";"))
	{ 
		errflg = 0;
		return lastst;
	} 

	if (match("#asm"))
	{ 
		doasm();
		lastst = STASM;
		return lastst;
	}

	/* if nothing else, assume it's an expression */
	doexpr();
	ns();
	lastst = STEXP;

	return lastst;
}

/*
** Semicolon enforcer
*/
/* called whenever syntax requires a semicolon */
void ns(void)
{
	if (!match(";"))
	{
		errrpt("Missing semicolon ';'.");
	}
	else
	{
		errflg = 0;
	}
}

/*
** evaluate an expression (allow , )
*/
void doexpr(void)
{
	do 
	{
		expression();

	} while (match(","));
}

/*
** Compound statement
*/
/* allow any number of statements to fall between "{}" */
void compound(void)
{
	short savsp = stkp; 
	short savloc = locptr;

	decl = 0;
	++ncmp;    /* new level open */

	while (!match("}"))
	{
		if (eof)
		{
			errrpt("Missing closing brace '}'."); 
			break;
		}
		else
		{
			statement(); /* do one */
		}
	}

	--ncmp;    /* close current level */

	if ((lastst != STRETURN) && (lastst != STGOTO))
	{
		modstk(savsp);
	}

	stkp = savsp;
	cptr = savloc;  /* release symtab space, retaining labels */

	while (cptr < locptr)
	{ 
		cptr2 = (short)(cptr + SYMSIZ);

		if (l_sym_tab[cptr + IDENT] == LABEL)
		{
			while (cptr < cptr2)
			{
				l_sym_tab[savloc++] = l_sym_tab[cptr++];
			}
		}
		else
		{
			cptr = cptr2;
		}
	}

	locptr = savloc;
	decl = -1;
}

/*
**   "if" statement
*/
void doif(void)
{
	short flab1 = getlabel();	/*	get label for false branch	*/
	short flab2 = 0;

	test(flab1, 1);				/*	get expression, and branch false	*/
	statement();				/*	if true, do a statement		*/

	if (!amatch("else", 4))		/*	if...else?					*/
	{ 
		/*	simple "if"...print false label	*/
		prlbll(flab1);
		return;		/*	and exit	*/
	}

	/*	an "if...else" statement.	*/
	flab2 = getlabel();

	if ((lastst != STRETURN) && (lastst != STGOTO))
	{
		jump(flab2);  /* bypass false code */
	}

	prlbll(flab1);  /* print false label */
	statement();    /* and do "else" clause */
	prlbll(flab2);  /* print true label */
}

/*
** "while" statement
*/
void dowhile(void)

{
	short lwq[WQSIZ] = { 0 };	/* allocate local queue */

	addwhile(lwq);			/*	add entry to queue	*/
							/*	(for "break" statement)	*/
	prlbll(lwq[WQLOOP]);	/*	loop label	*/

	if (!match("(1)"))
	{
		test(lwq[WQLAB], 1);	/*	if not while(1), see if true	*/
	}

	statement();			/*	if so, do a statement	*/
	jump(lwq[WQLOOP]);		/*	loop to label			*/
	prlbll(lwq[WQLAB]);		/*	exit label				*/
	delwhile();				/*	delete queue entry		*/
}

/*
** "do" statement
*/
void dodo(void)
{
	short lwq[WQSIZ] = { 0 };
	short top = 0;

	addwhile(lwq);
	top = getlabel();
	prlbll(top);

	statement();
	needbrack("while");

	prlbll(lwq[WQLOOP]);
	test(lwq[WQLAB], 1);
	jump(top);

	prlbll(lwq[WQLAB]);
	delwhile();
	ns();
}

/*
** "for" statement
*/
void dofor(void)
{
	short lwq[WQSIZ] = { 0 };
	short lab1 = 0;
	short lab2 = 0;

	addwhile(lwq);

	lab1 = getlabel();
	lab2 = getlabel();

	needbrack("(");

	if (!match(";"))
	{ 
		doexpr();    /* expr 1 */
		ns();
	}

	prlbll(lab1);

	if (!match(";"))
	{ 
		test(lwq[WQLAB], 0); /* expr 2 */
		ns();
	}

	jump(lab2);
	prlbll(lwq[WQLOOP]);

	if (!match(")"))
	{ 
		doexpr();      /* expr 3 */
		needbrack(")");
	}

	jump(lab1);
	prlbll(lab2);

	statement();

	jump(lwq[WQLOOP]);
	prlbll(lwq[WQLAB]);

	delwhile();
}

/*
** "switch" statement
*/
void doswitch(void)

{
	short lwq[WQSIZ] = { 0 };
	short endlab = 0;
	short swact = swactive;
	short swdef = swdefault;
	short* swnex = swnext;
	short* swptr = swnext;

	addwhile(lwq);

	wq[wqindex - 2] = 0;	/*	clear loop label	*/

	needbrack("(");
	doexpr();			/*	eval switch expression	*/
	needbrack(")");

	swdefault = 0;
	swactive = TRUE;
	endlab = getlabel();

	jump(endlab);
	statement(); /* cases etc. */
	jump(lwq[WQLAB]);
	prlbll(endlab);
	dcall("SWCH");

	while (swptr < swnext)		/* build dispatch table */
	{ 
		ddata();
		printlabel(*swptr++);	/* case label */
		outbyte(',');
		outdcl(*swptr++);		/* case value */
	}

	ddata();
	outdcl(0);

	if (swdefault)
	{
		jump(swdefault);
	}

	prlbll(lwq[WQLAB]);
	delwhile();

	swnext = swnex;
	swdefault = swdef;
	swactive = swact;
}

/*
** "case" statement
*/
void docase(void)
{
	short val = 0;

	if (!swactive)
	{
		oocerr();
	}

	if (swnext > SWEND)
	{ 
		oflerr();
		return;
	}

	*swnext = getlabel();

	prlbll(*swnext++);
	constx(&val);

	*swnext++ = val;

	needbrack(":");
}

/*
** "default" statement
*/
void dodefault(void)
{
	if (swactive)
	{ 
		if (swdefault)
		{
			multidef();
		}
	}
	else
	{
		oocerr();
	}

	needbrack(":");
	swdefault = getlabel();
	prlbll(swdefault);
}

/*
** "goto" statement
*/
void dogoto(void)
{
	if (nogo > 0)
	{
		oocerr(); /* illegal with block locals */
	}
	else
	{
		noloc = 1;
	}

	if (symname(ssname))
	{
		jump(addlbl());
	}
	else
	{
		illname();
	}

	ns();
}


/*
** check for "label";
*/
short dolabel(void)
{
	short savlpt = 0;

	blanks();
	savlpt = lptr;

	if (symname(ssname))
	{ 
		if (get_ch() == ':')
		{ 
			prlbll(addlbl());
			return TRUE;
		}
		else
		{
			lptr = savlpt;
		}
	}

	return FALSE;
}


/*
** "return" statement
*/
void doreturn(void)
{
	/* if not end of statement, get an expression */
	if (!endst())
	{
		doexpr();
	}

	dumplocsyms();

	modstk(0);  /* clean up stk */
	ret();    /* and exit function */
}

/*
** "break" statement
*/
void dobreak(void)
{
	short my_ptr = readwhile(wqindex);	/* see if any "whiles" are open */

	if (!my_ptr)
	{
		return;  /* no */
	}

	modstk((wq[my_ptr + WQSP]));  /* else clean up stk ptr */
	jump(wq[my_ptr + WQLAB]);  /* jump to exit label */
}

/*
** "continue" statement
*/
void docont(void)
{
	short my_ptr = wqindex;

	/* see if any "whiles" are open */
	for ( ; ; )
	{ 
		my_ptr = readwhile(my_ptr);

		if (!my_ptr)
		{
			return;  /* no */
		}

		if (wq[my_ptr + WQLOOP])
		{
			break;
		}
	}

	modstk((wq[my_ptr + WQSP]));  /* else clean up stk ptr */
	jump(wq[my_ptr + WQLOOP]);  /* jump to loop label */
}

/*
** "asm" pseudo-statement
*/
/* enters mode where assembly language statement are
** passed intact through parser */
void doasm(void)
{
	cmode = FALSE; /* mark mode as "asm" */

	for ( ; ; )
	{ 
		inline();  /* get and print lines */

		if (match("#endasm"))
		{
			break;  /* until... */
		}

		if (eof)
		{
			break;
		}

		outl(line);
	}

	kill(); /* invalidate line */
	cmode = TRUE; /* then back to parse level */
}